/* ml_codeconv.c - (multilingual) code conversion */
/*			by takada@seraph.ntt.jp   */

#ifdef JAPAN

#include "../h/mh.h"
#include <ctype.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

/* coding system */
typedef int coding_system_t;
#define CS_DEFAULT	 0
#define CS_JIS7		 1
#define CS_JEUC		 2
#define CS_SJIS		 3
#define CS_NOCONV	99

/* coding system list */
#define CSL_SIZE	3
#define CSL_DISPLAY(csl)	((csl)[0])
#define CSL_FILE(csl)		((csl)[1])
#define CSL_PROCESS(csl)	((csl)[2])

/* codeset status */
#define ASCII 		0
#define JISX0208	1

#define IS_JEUC(c)	((0xa1 <= (c)) && ((c) <= 0xfe))
#define IS_SJIS1(c)	(((0x81 <= (c)) && ((c) <= 0x9f)) || \
			 ((0xe0 <= (c)) && ((c) <= 0xfc)))
#define IS_SJIS2(c)	((0x40 <= (c)) && ((c) <= 0xfc) && ((c) != 0x7f))

/* hold coding system information */
static coding_system_t ml_coding_info[CSL_SIZE] =
       { CS_DEFAULT, CS_DEFAULT, CS_DEFAULT };

/* private functions */
static void read_profile();
static coding_system_t coding_to_cs();
static coding_system_t select_coding_system();
static void ml_fputs_sbr();
static void jeuc_fputs();
static void sjis_fputs();
static void jis7_fputs();
static void cntrl_putc();
static void ml_conv_sbr();


/*
 * Initialize: holding coding system information
 */
void
ml_init()
{
    read_profile(ml_coding_info);
}

/* get coding system list from profile and environment variable */
static void
read_profile(csl)
     coding_system_t csl[];
{
    char *s, *default_coding;

    if ((default_coding = getenv("LC_CTYPE")) == NULL)
      default_coding = getenv("LANG");

    if ((s = getenv("MH_DISPLAY_CODING")) == NULL)
      if ((s = m_find("display-coding")) == NULL)
	s = default_coding;
    CSL_DISPLAY(csl) = coding_to_cs(s);
    
    if ((s = getenv("MH_FILE_CODING")) == NULL)
      if ((s = m_find("file-coding")) == NULL)
	s = default_coding;
    CSL_FILE(csl) = coding_to_cs(s);

    if ((s = getenv("MH_PROCESS_CODING")) == NULL)
      if ((s = m_find("process-coding")) == NULL)
	s = default_coding;
    CSL_PROCESS(csl) = coding_to_cs(s);
}

static coding_system_t
coding_to_cs(coding)
     char *coding;
{
    if (uleq(coding,"C") || uleq(coding,"japanese")
	|| uleq(coding,"ja_JP.jis7") || uleq(coding,"ja_JP.jis8")
	|| uleq(coding,"ja_JP.pjis") || uleq(coding,"ja_JP.jis")
	|| uleq(coding,"wr_WR.ct")   || uleq(coding,"wr_WR.junet")) {
	return(CS_JIS7);
    } else if (uleq(coding,"ja_JP.ujis") || uleq(coding,"ja_JP.euc")) {
	return(CS_JEUC);
    } else if (uleq(coding,"ja_JP.mscode") || uleq(coding,"ja_JP.sjis")) {
	return(CS_SJIS);
    } else if (uleq(coding,"noconv")) {
	return(CS_NOCONV);
    } else {
	return(CS_DEFAULT);
    }
}

static coding_system_t
select_coding_system(stream)
     FILE *stream;
{
    struct stat buf;

    if (fstat(fileno(stream), &buf)) adios (NULLCP, "unable to fstat stream");
    switch (buf.st_mode & S_IFMT) {
      case S_IFREG:
	return(CSL_FILE(ml_coding_info));
      case S_IFIFO:
      case 0: /* some system returns zero-filled stat for pipe */
	return(CSL_PROCESS(ml_coding_info));
      case S_IFCHR:
      default:
	return(CSL_DISPLAY(ml_coding_info));
    }
}


/*
 *
 */
int
ml_ismlchar(c)
     unsigned char c;
{
    return(IS_JEUC(c));
}


/*
 * Output:
 */
void
ml_fputs(scanlk, stream)
     char *scanlk;
     FILE *stream;
{
    ml_fputs_sbr(scanlk, stream, 0);
}

void
ml_pretty_fputs(scanlk, stream)
     char *scanlk;
     FILE *stream;
{
    ml_fputs_sbr(scanlk, stream, 1);
}

static void
ml_fputs_sbr(scanlk, stream, pretty)
     char *scanlk;
     FILE *stream;
     int pretty;
{
    switch (select_coding_system(stream)) {
      case CS_NOCONV:
	fputs(scanlk, stream);
	break;
      case CS_SJIS:
	sjis_fputs(scanlk, stream, pretty);
	break;
      case CS_JEUC:
	jeuc_fputs(scanlk, stream, pretty);
	break;
      case CS_JIS7:
      case CS_DEFAULT:
      default:
	jis7_fputs(scanlk, stream, pretty);
	break;
    }
}


/*
 * Output routines with code conversion
 */
static void
jeuc_fputs(scanlk, stream, pretty)
     char *scanlk;
     FILE *stream;
     int pretty;
{
    unsigned char u1, u2;

    while (u1 = *scanlk++) {
	if (IS_JEUC(u1)) {
	    if (u2 = *scanlk++) {
		if (IS_JEUC(u2)) {
		    putc(u1, stream); putc(u2, stream);
		} else {
		    putc(' ', stream);
		    if (pretty && iscntrl(u2)) cntrl_putc(u2, stream);
		    else putc(u2, stream);
		}
	    } else {
		putc(u1, stream);
		break;
	    }
	} else if (pretty && iscntrl(u1)) {
	    cntrl_putc(u1, stream);
	} else {
	    putc(u1, stream);
	}
    }
}

#define E2S(i1, i2, o1, o2) {\
    (i1) &= ~0x80;\
    (i2) &= ~0x80;\
    (o1) = ((i1) - 0x21) / 2 + 0x81;\
    if ((o1) > 0x9f) { (o1) += (0xe0 - 0xa0); }\
    if ((i1) & 1) {\
	(o2) = (i2) + (0x40 - 0x21);\
	if ((o2) > 0x7e) (o2)++;\
    } else {\
	(o2) = (i2) + (0xfc - 0x7e);\
    }\
}

static void
sjis_fputs(scanlk, stream, pretty)
     char *scanlk;
     FILE *stream;
     int pretty;
{
    unsigned char u1, u2, s1, s2;

    while (u1 = *scanlk++) {
	if (IS_JEUC(u1)) {
	    if (u2 = *scanlk++) {
		if (IS_JEUC(u2)) {
		    E2S(u1, u2, s1, s2);
		    putc(s1, stream); putc(s2, stream);
		} else {
		    putc(' ', stream);
		    if (pretty && iscntrl(u2)) cntrl_putc(u2, stream);
		    else putc(u2, stream);
		}
	    } else {
		putc(u1, stream);
		break;
	    }
	} else if (pretty && iscntrl(u1)) {
	    cntrl_putc(u1, stream);
	} else {
	    putc(u1, stream);
	}
    }
}

#define DSGNT_JISX0208(stream, status) {\
    fputs("\033$B", (stream)); (status) = JISX0208;}\

#define DSGNT_ASCII(stream, status) {\
    fputs("\033(B", (stream)); (status) = ASCII;}\

static void
jis7_fputs(scanlk, stream, pretty)
     char *scanlk;
     FILE *stream;
     int pretty;
{
    int kanji_pos;	/* ASCII or JISX0208 */
    unsigned char u1, u2;

    kanji_pos = ASCII;
    while (u1 = *scanlk++) {
	if (IS_JEUC(u1)) {
	    if (u2 = *scanlk++) {
		if (IS_JEUC(u2)) {
		    if (kanji_pos == ASCII) DSGNT_JISX0208(stream, kanji_pos);
		    putc(u1 & 0x7f, stream); putc(u2 & 0x7f, stream);
		} else {
		    if (kanji_pos == JISX0208) DSGNT_ASCII(stream, kanji_pos);
		    putc(' ', stream);
		    if (pretty && iscntrl(u2)) cntrl_putc(u2, stream);
		    else putc(u2, stream);
		}
	    } else {
		if (kanji_pos == JISX0208) DSGNT_ASCII(stream, kanji_pos);
		putc(u1, stream);
		break;
	    }
	} else if (pretty && iscntrl(u1)) {
	    if (kanji_pos == JISX0208) DSGNT_ASCII(stream, kanji_pos);
	    cntrl_putc(u1, stream);
	} else {
	    if (kanji_pos == JISX0208) DSGNT_ASCII(stream, kanji_pos);
	    putc(u1, stream);
	}
    }
    if (kanji_pos == JISX0208) DSGNT_ASCII(stream, kanji_pos);
}

static void
cntrl_putc(c, stream)
     char c;
     FILE *stream;
{
    switch (c) {
      case '\b': putc('\\', stream); putc('b', stream); break;
      case '\f': putc('\\', stream); putc('f', stream); break;
      case '\n': putc('\\', stream); putc('n', stream); break;
      case '\r': putc('\\', stream); putc('r', stream); break;
      case '\t': putc('\\', stream); putc('t', stream); break;
      default:   putc('^', stream);  putc(c ^ 0x40, stream); break;
    }
}


/*
 * Input:
 */
char *
ml_conv(s)
     char *s;
{
    coding_system_t coding;
    
    if ((s == NULL) || ((coding = CSL_FILE(ml_coding_info)) == CS_NOCONV)) 
      return(s);
    
    ml_conv_sbr(s, coding);
    return(s);
}

#define ESC	'\033'

#define E2E(i1, i2, o1, o2) { (o1) = (i1); (o2) = (i2); }

#define I2E(i1, i2, o1, o2) { (o1) = ((i1) | 0x80); (o2) = ((i2) | 0x80); }

#define S2E(i1, i2, o1, o2) {\
    if ((i1) >= 0xe0) { (i1) -= (0xe0 - 0xa0); }\
    if ((i2) >= 0x80) { (i2)--; }\
    if ((i2) < 0x40 + 94) {\
	(o2) = (((((i1) - 0x81) * 2) + 0x21) | 0x80);\
	(o2) = (((i2) - (0x40 - 0x21)) | 0x80);\
    } else {\
	(o2) = (((((i1) - 0x81) * 2) + 0x21 + 1) | 0x80);\
	(o2) = (((i2) - (0x9e - 0x21)) | 0x80);\
    }\
}

static void
ml_conv_sbr(in, cs)
     char *in;
     coding_system_t cs;
{
    char *out = in;
    int kanji_pos = ASCII;
    unsigned char c1, c2, c3;
    
    while (c1 = *in++) {
	if (c1 == ESC) {
	    if (c2 = *in++) {
		if (c2 == '$') {
		    if (c3 = *in++) {
			if ((c3 == '@') || (c3 == 'B')) {
			    kanji_pos = JISX0208;
			} else {
			    *out++ = c1; *out++ = c2; *out++ = c3;
			}
		    } else {
			*out++ = c1; *out++ = c2;
			break;
		    }
		} else if (c2 == '(') {
		    if (c3 = *in++) {
			if ((c3 == 'B') || (c3 == 'J') || (c3 == 'H')) {
			    kanji_pos = ASCII;
			} else {
			    *out++ = c1; *out++ = c2; *out++ = c3;
			}
		    } else {
			*out++ = c1; *out++ = c2;
			break;
		    }
		} else {
		    *out++ = c1; *out++ = c2;
		}
	    } else {
		*out++ = c1;
		break;
	    }
	} else if ((cs == CS_SJIS) ? IS_SJIS1(c1) : IS_JEUC(c1)) {
	    if (c2 = *in++) {
		if ((cs == CS_SJIS) ? IS_SJIS2(c2) : IS_JEUC(c2)) {
		    if (cs == CS_SJIS) {
			S2E(c1, c2, *out++, *out++);
		    } else {
			E2E(c1, c2, *out++, *out++);
		    }
		} else {
		    *out++ = c1; *out++ = c2;
		}
	    } else {
		*out++ = c1;
		break;
	    }
	} else if (kanji_pos == JISX0208) {
	    if (c2 = *in++) {
		I2E(c1, c2, *out++, *out++);
	    } else {
		*out++ = c1;
		break;
	    }
	} else {
	    *out++ = c1;
	}
    }
    *out = 0;
}

#endif /* JAPAN */
